////////////////////////////////////////////////////////////////////// 
// extras/pedstats.cpp 
// (c) 2000-2001 Goncalo Abecasis
// 
// This file is distributed as part of the MERLIN source code package   
// and may not be redistributed in any form, without prior written    
// permission from the author. Permission is granted for you to       
// modify this file for your own personal use, but modified versions  
// must retain this copyright notice and must not be distributed.     
// 
// Permission is granted for you to use this file to compile MERLIN.    
// 
// All computer programs have bugs. Use this file at your own risk.   
// 
// Friday December 14, 2001
// 
 
#include "Constant.h"
#include "MathStats.h"
#include "MathMatrix.h"
#include "MathConstant.h"
#include "Parameters.h"
#include "Pedigree.h"
#include "Error.h"
#include "IBD.h"

#include <stdio.h>

void ReadParameters(int argc, char * argv[]);
void FamilyStatistics();
void RelativePairs(bool (*filter)(Person &, Person &) = NULL);
void TraitStatistics();
void CovarStatistics();
void AffectionStatistics();
void MarkerStatistics();
void HWStatistics();

int Generation(Person * p);
void MakeIndex(IntArray & index, IntArray & counts);

String pedName = "qtdt.ped";
String dataName = "qtdt.dat";
String ibdName = "qtdt.ibd";

Pedigree ped;
IBDTable ibd;

bool hardy = false;
bool nocheck = false;
bool pairs = false;

int main(int argc, char * argv[])
   {
   printf("Pedigree Statistics\n(c) 1999-2001 Goncalo Abecasis\n\n");

   ReadParameters(argc, argv);

   ped.Prepare(dataName);
   ped.Load(pedName);
   ibd.Load(ped, ibdName);

   FamilyStatistics();
   TraitStatistics();
   CovarStatistics();
   AffectionStatistics();

   if (!nocheck) ped.InheritanceCheck();

   MarkerStatistics();

   if (hardy) HWStatistics();

   return 0;
   };

int Generation(Person * p)
   {
   if (p->isFounder())
      return 0;

   int father = Generation(p->father);
   int mother = Generation(p->mother);

   return (mother > father ? mother : father) + 1;
   }

BEGIN_LONG_PARAMETERS(longParameters)
   LONG_PARAMETER("hardy-weinberg", &hardy)
   LONG_PARAMETER("ignore-mendelian-errors", &nocheck)
   LONG_PARAMETER("pairs", &pairs)
END_LONG_PARAMETERS();

void ReadParameters(int argc, char * argv[])
   {
   ParameterList pl;

   pl.Add(new StringParameter('p', "QTDT Pedigree File", pedName));
   pl.Add(new StringParameter('d', "QTDT Data File", dataName));
   pl.Add(new StringParameter('i', "QTDT IBD File", ibdName));
   pl.Add(new StringParameter('x', "Missing Value Code", Pedigree::missing));
   pl.Add(new LongParameters("Additional Options", longParameters));

   pl.Read(argc, argv);
   pl.Status();
   }

void FamilyStatistics()
   {
   if (ped.familyCount == 0)
      {
      printf("WARNING - Pedigree is empty");
      return;
      }

   int families = 0, founders = 0, nonfounders = 0;
   int maxgen = 0, mingen = 9999;
   int maxcount, mincount;
   double avecount = 0.0, avegen = 0.0;

   maxcount = mincount = ped.families[0]->count;

   IntArray generations(ped.familyCount);

   for (int f = 0; f < ped.familyCount; f++)
      {
      families++;
      founders    += ped.families[f]->founders;
      nonfounders += ped.families[f]->count - ped.families[f]->founders;

      maxcount = max(ped.families[f]->count, maxcount);
      mincount = min(ped.families[f]->count, mincount);
      avecount += ped.families[f]->count;

      generations[f] = 0;
      for (int i = ped.families[f]->first; i <= ped.families[f]->last; i++)
         generations[f] = max(Generation(ped.persons[i]) + 1, generations[f]);

      maxgen = max(generations[f], maxgen);
      mingen = min(generations[f], mingen);
      avegen += generations[f];
      }

   avegen /= (double) max(families, 1);
   avecount /= (double) max(families, 1);

   IntArray counts(maxcount + 1), gens(maxgen + 1);

   counts.Set(0);
   gens.Set(0);

   for (int f = 0; f < ped.familyCount; f++)
      {
      counts[ped.families[f]->count]++;
      gens[generations[f]]++;
      }

   IntArray idxcount(maxcount + 1), idxgen(maxgen + 1);

   MakeIndex(idxcount, counts);
   MakeIndex(idxgen, gens);

   double fdiv = max((double) families, 1.0) / (double) 100.0;

   printf("PEDIGREE STRUCTURE\n"
          "==================\n"
          "  Individuals: %d (%d founders, %d nonfounders)\n"
          "     Families: %d\n\n"
          " Family Sizes\n"
          "      Average: %.2f (%d to %d)\n"
          " Distribution: %d (%.1f%%), %d (%.1f%%) and %d (%.1f%%)\n\n"
          "  Generations\n"
          "      Average: %.2f (%d to %d)\n"
          " Distribution: %d (%.1f%%), %d (%.1f%%) and %d (%.1f%%)\n\n",
          founders + nonfounders, founders, nonfounders, families,
          avecount, mincount, maxcount,
          idxcount[0], (double) counts[idxcount[0]] / fdiv,
          idxcount[1], (double) counts[idxcount[1]] / fdiv,
          idxcount[2], (double) counts[idxcount[2]] / fdiv,
          avegen, mingen, maxgen,
          idxgen[0], (double) gens[idxgen[0]] / fdiv,
          idxgen[1], (double) gens[idxgen[1]] / fdiv,
          idxgen[2], (double) gens[idxgen[2]] / fdiv);

   if (pairs) RelativePairs();
   }

int trait_filter = 0;
bool FilterOnTrait(Person & p1, Person & p2)
   {
   return ((p1.traits[trait_filter] != _NAN_) &&
           (p2.traits[trait_filter] != _NAN_));
   }

void TraitStatistics()
   {
   if (Person::traitCount == 0) return;

   int count = 0, possible = 0, fcount = 0, fpossible = 0;

   printf("QUANTITATIVE TRAIT STATISTICS\n"
          "=============================\n"
          "%15s %15s %15s %10s %10s\n",
          "", "[Phenotypes]", "[Founders]", "Mean", "Var");

   for (int t = 0; t < Person::traitCount; t++)
     {
     int    tcount = 0, tpossible = 0, tfcount = 0, tfpossible = 0;
     double mean = 0, var = 0;

     for (int i = 0; i < ped.count; i++)
         {
         tpossible  ++;
         tfpossible += ped[i].isFounder();
         if ( !ped[i].isPhenotyped(t) ) continue;
         mean    += ped[i].traits[t];
         var     += ped[i].traits[t] * ped[i].traits[t];
         tfcount += ped[i].isFounder();
         tcount  ++;
         }

     possible  += tpossible;
     fpossible += tfpossible;
     count     += tcount;
     fcount    += tfcount;

     var = (tcount <= 1) ? 0.0 :
            (tcount * var - mean * mean) /
           ((tcount) * (tcount - 1));
     mean /= max(tcount, 1);

     printf("%15.15s %8d %5.1f%% %8d %5.1f%% %10.3f %10.3f\n",
            (const char *) Person::traitNames[t],
            tcount, (double) tcount / (double)  max(tpossible, 1) * 100,
            tfcount, (double) tfcount / (double) max(tfpossible, 1) * 100,
            mean, var);
     }
   printf("%15s %8d %5.1f%% %8d %5.1f%%\n\n", "Total",
           count,  (double) count  / (double) max(possible, 1) * 100,
           fcount, (double) fcount / (double) max(fpossible, 1) * 100 );

   if (pairs)
      {
      printf("%15s %12s %5s %9s %11s %10s %8s\n",
              "PairCounts", "ParentChild", "Sibs", "HalfSibs",
              "Grandparent", "Avuncular", "Cousins");

      for (trait_filter = 0; trait_filter < ped.traitCount; trait_filter++)
         {
         printf("%15.15s ", (const char *) ped.traitNames[trait_filter]);
         RelativePairs(FilterOnTrait);
         }
      printf("\n");
      }
   }

void CovarStatistics()
   {
   if (Person::covariateCount == 0) return;

   int count = 0, possible = 0, fcount = 0, fpossible = 0;

   printf("COVARIATE STATISTICS\n"
          "====================\n"
          "%15s %15s %15s %10s %10s\n",
          "", "[Covariates]", "[Founders]", "Mean", "Var");

   for (int c = 0; c < Person::covariateCount; c++)
     {
     int    tcount = 0, tpossible = 0, tfcount = 0, tfpossible = 0;
     double mean = 0, var = 0;

     for (int i = 0; i < ped.count; i++)
         {
         tpossible  ++;
         tfpossible += ped[i].isFounder();
         if ( !ped[i].isControlled(c) ) continue;
         mean    += ped[i].covariates[c];
         var     += ped[i].covariates[c] * ped[i].covariates[c];
         tfcount += ped[i].isFounder();
         tcount  ++;
         }

     possible  += tpossible;
     fpossible += tfpossible;
     count     += tcount;
     fcount    += tfcount;

     var = (tcount <= 1) ? 0.0 :
            (tcount * var - mean * mean) /
           ((tcount) * (tcount - 1));
     mean /= max(tcount, 1);

     printf("%15.15s %8d %5.1f%% %8d %5.1f%% %10.3f %10.3f\n",
            (const char *) Person::covariateNames[c],
            tcount, (double) tcount / (double)  max(tpossible, 1) * 100,
            tfcount, (double) tfcount / (double) max(tfpossible, 1) * 100,
            mean, var);
     }
   printf("%15s %8d %5.1f%% %8d %5.1f%%\n\n", "Total",
           count,  (double) count  / (double) max(possible, 1) * 100,
           fcount, (double) fcount / (double) max(fpossible, 1) * 100 );
   }

int affection_filter = 0;
bool FilterOnAffection(Person & p1, Person & p2)
   {
   return ((p1.affections[affection_filter] != _NAN_) &&
           (p2.affections[affection_filter] != _NAN_));
   }

void AffectionStatistics()
   {
   if (Person::affectionCount == 0) return;

   int count = 0, possible = 0, fcount = 0, fpossible = 0;

   printf("AFFECTION STATISTICS\n"
          "====================\n"
          "%15s %15s %15s %10s\n",
          "", "[Diagnostics]", "[Founders]", "Prevalence");

   for (int a = 0; a < Person::affectionCount; a++)
     {
     int    tcount = 0, tpossible = 0, tfcount = 0, tfpossible = 0;
     int    affecteds = 0;

     for (int i = 0; i < ped.count; i++)
         {
         tpossible  ++;
         tfpossible += ped[i].isFounder();
         if ( !ped[i].isDiagnosed(a) ) continue;
         affecteds += ped[i].affections[a] == 2;
         tfcount   += ped[i].isFounder();
         tcount    ++;
         }

     possible  += tpossible;
     fpossible += tfpossible;
     count     += tcount;
     fcount    += tfcount;

     printf("%15.15s %8d %5.1f%% %8d %5.1f%% %9.1f%%\n",
            (const char *) Person::affectionNames[a],
            tcount, (double) tcount / (double)  max(tpossible, 1) * 100,
            tfcount, (double) tfcount / (double) max(tfpossible, 1) * 100,
            (double) affecteds / (double) max(tcount, 1) * 100 );
     }
   printf("%15s %8d %5.1f%% %8d %5.1f%%\n\n", "Total",
           count,  (double) count  / (double) max(possible, 1) * 100,
           fcount, (double) fcount / (double) max(fpossible, 1) * 100 );

   if (pairs)
      {
      printf("%15s %12s %5s %9s %11s %10s %8s\n",
              "PairCounts", "ParentChild", "Sibs", "HalfSibs",
              "Grandparent", "Avuncular", "Cousins");

      for (affection_filter = 0; affection_filter < ped.affectionCount; affection_filter++)
         {
         printf("%15.15s ", (const char *) ped.affectionNames[affection_filter]);
         RelativePairs(FilterOnAffection);
         }
      printf("\n");
      }
   }

void MarkerStatistics()
   {
   if (Person::markerCount == 0) return;

   printf("MARKER GENOTYPE STATISTICS\n"
          "==========================\n"
          "%15s %15s %15s %10s %8s\n",
          "", "[Genotypes]", "[Founders]", "Hetero", "IBD");

   int genotypes = 0, possible = 0, fgenotypes = 0, fpossible = 0,
       hetero = 0;

   for (int m = 0; m < Person::markerCount; m++)
     {
     int mGenotypes = 0, mPossible = 0, mfGenotypes = 0, mfPossible = 0,
         mHetero = 0, lastAllele = 0;

     for (int i = 0; i < ped.count; i++)
         {
         mPossible  ++;
         mfPossible += ped[i].isFounder();
         if ( !ped[i].isGenotyped(m) ) continue;
         mGenotypes  ++;
         mfGenotypes += ped[i].isFounder();
         mHetero     += ped[i].markers[m].isHeterozygous();

         lastAllele = max( max(ped[i].markers[m][0], ped[i].markers[m][1]),
                           lastAllele);
         }

     possible   += mPossible;
     fpossible  += mfPossible;
     fgenotypes += mfGenotypes;
     genotypes  += mGenotypes;
     hetero     += mHetero;

     int ibdCount = 0;
     if (!ibd.isEmpty())
        for (int f = 0; f < ped.familyCount; f++)
            if (ibd.HaveFamily(m, ped.families[f]))
                ibdCount++;

     printf("%15.15s %8d %5.1f%% %8d %5.1f%% %9.1f%% %6d/%d families\n",
            (const char *) Person::markerNames[m],
            mGenotypes, (double) mGenotypes / (double) max(mPossible, 1) * 100,
            mfGenotypes, (double)mfGenotypes / (double)max(mfPossible, 1) * 100,
            (double) mHetero / (double) max(mGenotypes, 1) * 100,
            ibdCount, ped.familyCount);
     }

     printf("%15s %8d %5.1f%% %8d %5.1f%% %9.1f%%\n\n", "Total",
            genotypes, (double) genotypes / (double) max(possible, 1) * 100,
            fgenotypes, (double) fgenotypes / (double) max(fpossible, 1) * 100,
            (double) hetero / (double) max(genotypes, 1) * 100);
   }

void MakeIndex(IntArray & index, IntArray & counts)
   {
   index.Dimension(counts.Length());
   index.SetSequence(0, 1);

   // Insertion sort should be OK, for N < 20
   for (int i, j = 1; j < counts.Length(); j++)
      {
      int tmp = counts[j];    // index[j] = j
      for (i = j - 1; (i >= 0) && (counts[index[i]] < tmp); i--)
         index[i + 1] = index[i];
      index[i + 1] = j;
      }

   // Make sure we have at least three items
   for (int i = counts.Length() + 1; i < 4; i++)
      {
      counts.Dimension(i);
      index.Dimension(i);

      counts[i - 1] = 0;
      index[i - 1] = i - 1;
      }
   }

void HWStatistics()
   {
   if (ped.markerCount == 0) return;

   bool fail = false;
   bool note = false;

   printf("HARDY-WEINBERG CHECK\n"
          "====================\n");

   Matrix observed;
   Vector freqs;

   for (int m = 0; m < Person::markerCount; m++)
      {
      int alleleCount = ped.CountAlleles(m);

      if (alleleCount <= 1) continue;

      observed.Dimension(alleleCount, alleleCount);
      observed.Zero();

      freqs.Dimension(alleleCount);
      freqs.Zero();

      int total = 0;

      for (int i = 0; i < ped.count; i++)
         if (ped[i].isGenotyped(m))
            {
            int lo = ped[i].markers[m].Lo() - 1;
            int hi = ped[i].markers[m].Hi() - 1;

            observed[hi][lo]++;
            freqs[hi]++;
            freqs[lo]++;
            total++;
            }

      freqs.Multiply(1.0 / (2.0 * total));

      double chisq = 0.0;
      bool   low_counts = false;

      for (int i = 0; i < alleleCount; i++)
         if (freqs[i] > 0)
            for (int j = 0; j <= i; j++)
               if (freqs[j] > 0)
                  {
                  double expected = freqs[i] * freqs[j] * total * (i == j ? 1 : 2);
                  chisq += square(expected - observed[i][j]) / expected;
                  low_counts |= expected < 3.0;
                  }

      note |= low_counts;
      int realAlleles = 0;

      for (int i = 0; i < alleleCount; i++)
         if (freqs[i] != 0.0)
            realAlleles++;

      double pvalue = realAlleles == 1.0 ? 1.0 :
                      chidist(chisq, realAlleles * (realAlleles - 1) * 0.5);

      if (pvalue > 0.05) continue;

      if (!fail)
         printf("%15s %10s %10s %10s %10s\n",
                "", "Alleles", "Largest", "ChiSq", "P-value");

      fail = true;

      printf("%15.15s %10d %10d %10.2f %10.4f %s\n",
             (const char *) Person::markerNames[m],
             alleleCount, realAlleles, chisq, pvalue,
             low_counts ? "see note" : "");
     }

   if (!fail)
      printf("All markers okay.\n");

   if (note && fail)
      printf("\nNOTE: Some genotypes had very low expecteds counts,\n"
             "      making this HW test unreliable.\n");
   }

bool isParent(Person * a, Person * b)
   {
   Person * ancestor = (a->traverse < b->traverse) ? a : b;
   Person * descendant = (a->traverse < b->traverse) ? b : a;

   return (descendant->father == ancestor || descendant->mother == ancestor);
   }

bool isGrandParentOffspring(Person * a, Person * b)
   {
   Person * ancestor = (a->traverse < b->traverse) ? a : b;
   Person * descendant = (a->traverse < b->traverse) ? b : a;

   if (descendant->isFounder()) return false;

   if (!descendant->father->isFounder() &&
       descendant->father->father == ancestor ||
       descendant->father->mother == ancestor)
       return true;

   if (!descendant->mother->isFounder() &&
       descendant->mother->father == ancestor ||
       descendant->mother->mother == ancestor)
       return true;

   return false;
   }

bool isAvuncular(Person * a, Person * b)
   {
   if (!a->isFounder())
      if (b->isSib(*a->father) || b->isSib(*a->mother))
         return true;

   if (!b->isFounder())
      if (a->isSib(*b->father) || a->isSib(*b->mother))
         return true;

   return false;
   }

bool isCousin(Person *a, Person *b)
   {
   if (a->isFounder() || b->isFounder() || a->isSib(*b))
      return false;

   if (!a->mother->isFounder())
      if (!b->mother->isFounder() && a->mother->isSib(*b->mother) ||
          !b->father->isFounder() && a->mother->isSib(*b->father) )
         return true;

   if (!a->father->isFounder())
      if (!b->mother->isFounder() && a->father->isSib(*b->mother) ||
          !b->father->isFounder() && a->father->isSib(*b->father) )
         return true;

   return false;
   }

void RelativePairs(bool (*filter)(Person &, Person &))
   {
   int parents = 0, sibs = 0, halfsibs = 0;
   int uncles = 0, cousins = 0, grandparents = 0;

   for (int f = 0; f < ped.familyCount; f++)
      for (int i = ped.families[f]->first; i <= ped.families[f]->last; i++)
         for (int j = i + 1; j <= ped.families[f]->last; j++)
            if (filter == NULL || filter(ped[i], ped[j]))
               if (isParent(&ped[i], &ped[j])) parents++;
               else if (ped[i].isSib(ped[j])) sibs++;
               else if (ped[i].isHalfSib(ped[j])) halfsibs++;
               else if (isAvuncular(&ped[i], &ped[j])) uncles++;
               else if (isCousin(&ped[i], &ped[j])) cousins++;
               else if (isGrandParentOffspring(&ped[i], &ped[j])) grandparents++;

   if (filter != NULL)
      {
      printf("%12d %5d %9d %11d %10d %8d\n",
             parents, sibs, halfsibs, grandparents, uncles, cousins);
      return;
      }

   if (parents == 0)
      printf(" No relative pairs\n\n");
   else
      {
      printf(" Relative Pair Counts\n"
             "        Parent-Child: %d pairs\n", parents);
      if (sibs) printf("           Sib-pairs: %d pairs\n", sibs);
      if (halfsibs) printf("           Half-Sibs: %d pairs\n", halfsibs);
      if (grandparents) printf("  Grandparent-Nephew: %d pairs\n", grandparents);
      if (uncles) printf("           Avuncular: %d pairs\n", uncles);
      if (cousins) printf("             Cousins: %d pairs\n", cousins);
      printf("\n");
      }
   }
 
