////////////////////////////////////////////////////////////////////// 
// merlin/MerlinFamily.cpp 
// (c) 2000-2001 Goncalo Abecasis
// 
// This file is distributed as part of the MERLIN source code package   
// and may not be redistributed in any form, without prior written    
// permission from the author. Permission is granted for you to       
// modify this file for your own personal use, but modified versions  
// must retain this copyright notice and must not be distributed.     
// 
// Permission is granted for you to use this file to compile MERLIN.    
// 
// All computer programs have bugs. Use this file at your own risk.   
// 
// Friday December 14, 2001
// 
 
#include "MerlinFamily.h"
#include "MerlinHaplotype.h"
#include "MapFunction.h"
#include "MathConstant.h"
#include "NPL-ASP.h"
#include "NPL-QTL.h"

#include <math.h>
#include <stdarg.h>

// What size families are we analysing
int FamilyAnalysis::maxBits = 24;

// Positions to analyse
int FamilyAnalysis::steps_per_interval = 0;
double FamilyAnalysis::maxDistance = _NAN_;
double FamilyAnalysis::minDistance = _NAN_;
double FamilyAnalysis::grid  = _NAN_;
double FamilyAnalysis::start = _NAN_;
double FamilyAnalysis::stop  = _NAN_;

// General analyses
bool FamilyAnalysis::findErrors = false;
bool FamilyAnalysis::estimateIBD = false;
bool FamilyAnalysis::estimateKinship = false;
bool FamilyAnalysis::estimateInformation = false;
bool FamilyAnalysis::estimateMatrices = false;
bool FamilyAnalysis::storeKinship = false;

// Haplotyping
bool FamilyAnalysis::bestHaplotype = false;
int  FamilyAnalysis::sampledHaplotypes = 0;
bool FamilyAnalysis::allHaplotypes = false;

// Non-parametric linkage analysis
bool FamilyAnalysis::nplAll = false;
bool FamilyAnalysis::nplPairs = false;
bool FamilyAnalysis::nplQtl = false;
bool FamilyAnalysis::nplDeviates = false;

// Output options
bool FamilyAnalysis::quietOutput = false;
bool FamilyAnalysis::useMarkerNames = false;
bool FamilyAnalysis::simwalk2 = false;

// Swap options
bool FamilyAnalysis::useSwap = false;

// Internal flags to minimize useless calculations
bool FamilyAnalysis::multipoint = false;
bool FamilyAnalysis::scanChromosome = false;
bool FamilyAnalysis::betweenMarkers = false;

// Flags for controlling recombination
bool FamilyAnalysis::zeroRecombination = false;
bool FamilyAnalysis::oneRecombination = false;
bool FamilyAnalysis::twoRecombinations = false;
bool FamilyAnalysis::threeRecombinations = false;
bool FamilyAnalysis::twopoint = false;

FamilyAnalysis::FamilyAnalysis(Pedigree & pedigree) :
   mantra(), kinship(mantra), ibd(mantra), matrix(mantra),
   kac(labels), ped(pedigree), hybrid(*this)
   {
   lowBound = pow(2.0, -129);
   rescale  = pow(2.0, 258);
   printHeader = nplCount = 0;
   npl = NULL;
   }

FamilyAnalysis::~FamilyAnalysis()
   {
   if (npl != NULL) delete [] npl;
   }

bool FamilyAnalysis::SelectFamily(Family * f)
   {
   mantra.Prepare(ped, *f);
   mantra.PrepareIBD();

   printHeader = true;
   cleanOutput = false;

   family = f;
   serial = f->serial;

   if (estimateMatrices) matrix.SelectFamily(&ped, family);
   if (estimateIBD) ibd.SelectFamily(&ped, family);
   if (estimateKinship || storeKinship) kinship.SelectFamily(&ped, family);

   // if (mantra.couples)
   //   PrintMessage("  %d symmetric founder couple%s detected",
   //               mantra.couples, mantra.couples == 1 ? "" : "s");

   if (mantra.bit_count <= maxBits)
      return true;

   // Tell simwalk2 we have skipped this family
   if (simwalk2) hybrid.Skip();

   PrintMessage("  SKIPPED: Maximum complexity set at %d bits", maxBits);
   CleanMessages();
   return false;
   }

void FamilyAnalysis::SetupGlobals()
   {
   // How many NPL statistics to calculate
   nplCount = (nplAll + nplPairs) * ped.affectionCount +
              (nplQtl + nplDeviates) * ped.traitCount;

   if (npl != NULL) delete [] npl;

   nplInv.Dimension(nplCount);
   nplMean.Dimension(nplCount);
   npl = new Tree[nplCount];

   // Carry out between marker analyses?
   betweenMarkers = estimateIBD || estimateKinship || estimateInformation ||
                    estimateMatrices || storeKinship || nplCount != 0;

   // Carry out analyses at each marker?
   scanChromosome = betweenMarkers || findErrors;

   // Perform right conditional calculations?
   multipoint     = scanChromosome || bestHaplotype && zeroRecombination ||
                    (sampledHaplotypes != 0) || allHaplotypes;

   // Set the maximum number of recombinants in multipoint calculations
   Multipoint::maximum_recombinants =
      threeRecombinations * 3 + twoRecombinations * 2 + oneRecombination;

   // Set up labels for NPL output
   kac.LabelPhenotypes(ped, nplAll, nplPairs, nplQtl, nplDeviates);
   }

int FamilyAnalysis::SetupMap(int chromosome)
   {
   markers.Dimension(0);
   int next = PedigreeGlobals::SortMarkersInMapOrder(markers, chromosome);

   markerCount = markers.Length();

   if (markerCount == 0) error("No markers could be placed");

   // Clear previous information
   labels.Clear();
   analysisPositions.Clear();
   markerPositions.Clear();

   String label;

   // Retrieve marker positions
   for (int i = 0; i < markerCount; i++)
      {
      MarkerInfo * info = PedigreeGlobals::GetMarkerInfo(markers[i]);
      ValidateMarker(info);
      markerPositions.Push(info->position);
      }

   // For maps with no recombination, only analyse one location
   if (zeroRecombination)
      {
      analysisPositions.Push(markerPositions[0]);
      labels.Push(useMarkerNames ?
                  PedigreeGlobals::GetMarkerInfo(markers[0])->name :
                  label = (markerPositions[0] * 100.0));
      }
   // Singlepoint analysis at marker locations only
   else if (twopoint)
      {
      analysisPositions = markerPositions;

      for (int i = 0; i < markerCount; i++)
         labels.Push(PedigreeGlobals::GetMarkerInfo(markers[0])->name);
      }
   // Analysis along grid of equally spaced locations
   else if (grid != _NAN_)
      {
      double position = start != _NAN_ ? start * 0.01 :
         PedigreeGlobals::GetMarkerInfo(markers[0])->position;
      double mapEnd = stop != _NAN_ ? stop * 0.01 :
         PedigreeGlobals::GetMarkerInfo(markers.Last())->position;
      double step = grid * 0.01;

      while (true)
         {
         analysisPositions.Push(position);
         labels.Push(label = position * 100.0);
         if (mapEnd - position < 1e-5) break;
         position += step;
         }
      }
   // Analysis at marker positions and equally spaced locations in-between
   else
      {
      double min = minDistance == _NAN_ ? _NAN_ : minDistance * 0.01;
      double max = maxDistance == _NAN_ ? _NAN_ : maxDistance * 0.01;

      double last = markerPositions[0];

      for (int i = 0; i < markerCount; i++)
         {
         if (markerPositions[i] > last)
            {
            double delta = markerPositions[i] - last;
            int    steps = steps_per_interval;
            double step  = delta / (steps + 1);

            if ((max != _NAN_) && (max < step))
               {
               steps = (int) (delta / (max + 1e-6));
               step  = delta / (steps + 1);
               }

            if ((min != _NAN_) && min > step)
               steps = (int) (delta / (minDistance  + 1e-6) - 0.999);

            if (steps < 0) steps = 0;

            for (int j = 1; j <= steps; j++)
               {
               double position = last + delta * j / (steps+1);

               analysisPositions.Push(position);
               labels.Push(label = position * 100.0);
               }
            }

         if (markerPositions[i] > last || i == 0)
            {
            analysisPositions.Push(last = markerPositions[i]);
            labels.Push(useMarkerNames ?
               PedigreeGlobals::GetMarkerInfo(markers[i])->name :
               label = (last * 100.0));
            }
         }
      }
   information.Dimension(2, analysisPositions.dim);
   information.Zero();

   if (nplCount != 0) kac.Prepare(nplCount, ped.familyCount);
   if (storeKinship) kinship.SetFamilyCount(ped.familyCount);

   return next;
   }

void FamilyAnalysis::ValidateMarker(MarkerInfo * info)
   {
#ifdef __USE_LONG_INT
   if (info->freq.dim > 64)
       error("Please recode marker %s (and possibly others), which has %d alleles\n"
             "This version of merlin supports up to 64 alleles",
             (const char *) info->name, info->freq.dim - 1);
#else
   if (info->freq.dim > 32)
       error("Please recode marker %s (and possibly others), which has %d alleles\n"
             "This version of merlin supports up to 32 alleles",
             (const char *) info->name, info->freq.dim - 1);
#endif
   }

void FamilyAnalysis::Analyse()
   {
   singlepoint = NULL;
   right = NULL;

   try
      {
      BasicTree::FreeSwap();

      ScoreSinglepoint();

      if (informativeCount == 0)
         SkipFamily();
      else
         {
         if (storeKinship) kinship.SetPositionCount(analysisPositions.dim);
         if (nplCount) ScoreNPL();
         if (multipoint)
            {
            if (!zeroRecombination && ScoreConditionals() ||
                 zeroRecombination && ZeroRecombination())
               {
               if (scanChromosome) ScanChromosome();
               if (allHaplotypes) haplo.All(*this);
               for (int i = 0; i < sampledHaplotypes; i++) haplo.Sample(*this);
               if (bestHaplotype && zeroRecombination) haplo.MostLikely(*this);
               }
            else
               PrintMessage("  SKIPPED: Requires impossible recombination pattern");
            if (!twopoint) delete [] right;
            if (simwalk2) hybrid.Output();
            }
         if (bestHaplotype && !zeroRecombination) haplo.MostLikely(*this);
         }

      CleanMessages();
      delete [] singlepoint;
      }
   catch (const TreesTooBig & problem)
      {
      PrintMessage("  SKIPPED: At least %d megabytes needed", problem.memory_request);
      CleanMessages();

      if (singlepoint != NULL) delete [] singlepoint;
      if (right != NULL) delete [] right;

      if (simwalk2) hybrid.Skip();
      if (nplCount != 0) kac.SkipFamily(family->serial);
      if (storeKinship) kinship.Discard();
      };
   };

void FamilyAnalysis::SkipFamily()
   {
   // If there is no information on this family
   // IBD and kinship coefficients / matrices are the same at all markers
   if (estimateIBD || estimateKinship || estimateMatrices)
      {
      double sum;
      Tree dummy;

      dummy.MakeMinimalTree(1.0, mantra.bit_count);

      double scale = pow(2.0, mantra.bit_count);

      if (estimateIBD) sum = ibd.Calculate(dummy, labels[0]) * scale;
      if (estimateKinship) sum = kinship.Calculate(dummy, labels[0]) * scale;
      if (estimateMatrices) sum = matrix.Calculate(dummy, labels[0]) * scale;

      for (int i = 1; i < labels.Length(); i++)
         {
         if (estimateIBD) ibd.Output(labels[i], sum);
         if (estimateKinship) kinship.Output(labels[i], sum);
         if (estimateMatrices) matrix.Output(labels[i], sum);
         }
      }

   // Output dummy haplotype -- something to look at -- keeps people happy
   if (allHaplotypes || bestHaplotype) haplo.UninformativeFamily(*this);

   for (int i = 0; i < sampledHaplotypes; i++)
      haplo.UninformativeFamily(*this);

   // And tell Simwalk2 this family isn't very interesting (to us!)
   if (simwalk2) hybrid.Skip();
   }

void FamilyAnalysis::ScoreNPL()
   {
   int         index = 0;

   if (nplAll)
      {
      NPL_ALL engine;

      for (int a = 0; a < ped.affectionCount; a++)
         {
         ProgressReport("Scoring NPL-ALL statistic", a, ped.affectionCount);

         engine.ScoreNPL(mantra, a);
         npl[index++].Copy(engine);
         }
      }

   if (nplPairs)
      {
      NPL_Pairs engine;

      for (int a = 0; a < ped.affectionCount; a++)
         {
         ProgressReport("Scoring NPL-Pairs statistic", a, ped.affectionCount);

         engine.ScoreNPL(mantra, a);
         npl[index++].Copy(engine);
         }
      }

   if (nplQtl)
      {
      NPL_QTL engine;

      for (int t = 0; t < ped.traitCount; t++)
         {
         ProgressReport("Scoring NPL-QTL statistic", t, ped.traitCount);

         engine.ScoreWithSampleMean(mantra, t);
         npl[index++].Copy(engine);
         }
      }

   if (nplDeviates)
      {
      NPL_QTL engine;

      for (int t = 0; t < ped.traitCount; t++)
         {
         ProgressReport("Scoring NPL-Deviate statistic", t, ped.traitCount);

         engine.ScoreWithMeanZero(mantra, t);
         npl[index++].Copy(engine);
         }
      }

   while (index--)
      {
      stats.GetBounds(npl[index]);

      nplMean[index] = stats.mean;
      if (stats.var > TINY)
         {
         nplInv[index] = 1.0 / sqrt(stats.var);
         kac.min[index][serial] = (stats.min - stats.mean) * nplInv[index];
         kac.max[index][serial] = (stats.max - stats.mean) * nplInv[index];
         }
      else
         nplInv[index] = 0.0;
      }
   }

void FamilyAnalysis::ScoreSinglepoint()
   {
   singlepoint = new InheritanceTree[ped.markerCount];
   InheritanceTree tempVectors;

   informativeCount = 0;
   informativeMarkers.Clear();
   informativePositions.Clear();
   informationScores.Clear();
   informationBounds.Clear();

#ifdef __PRINTOUT_STATS__
   Vector nodeCounts, leafCounts, zeroCounts, infoScores;
#endif

   for (int m = 0; m < markerCount; m++)
      {
      ProgressReport("Singlepoint", m, markerCount);
      tempVectors.ScoreVectors(mantra, markers[m], !allHaplotypes);

#ifndef __PRINTOUT_STATS__
      stats.GetTreeInfo(tempVectors);
#else
      stats.GetDetailedInfo(tempVectors);
      nodeCounts.Push(stats.nodeCount);
      leafCounts.Push(stats.leafCount);
      zeroCounts.Push(stats.zeroCount);
      infoScores.Push(stats.information);
#endif

      if (stats.mean == 0.0)
         {
         // TODO -- be helpful and try to pin down problem genotypes
         PrintMessage("  Skipping Marker %-10s [BAD INHERITANCE]",
              (const char *) ped.markerNames[markers[m]]);

         // Replace with uninformative inheritance tree
         tempVectors.MakeMinimalTree(1.0, mantra.bit_count);
         stats.information = 0.0;

         // Erase genotypes for this marker
         for (int i = family->first; i <= family->last; i++)
            ped[i].markers[markers[m]][2] = ped[i].markers[markers[m]][1] = 0;
         }

      if (stats.information > 1e-5 || twopoint)
         {
         singlepoint[informativeCount].Copy(tempVectors);
         singlepoint[informativeCount++].Pack();
         informativeMarkers.Push(markers[m]);
         informativePositions.Push(markerPositions[m]);
         informationScores.Push(stats.information);
         informationBounds.Push(stats.information > 0.3 ? stats.GetMinInfo(tempVectors) : 0.0);
         }
      else if (stats.information > 1e-10)
         PrintMessage("  Skipping Marker %-10s [Info = %.6f]",
              (const char *) ped.markerNames[markers[m]], stats.information);
      }

   if (Multipoint::maximum_recombinants || twopoint || zeroRecombination ||
       mantra.bit_count < 3)
      useSparse = false;
   else
      {
      // By default we don't assume sparseness
      useSparse = false;

      // Lower information bounds if founder couple symmetries are being used
      if (mantra.couples)
         // TODO -- work out what a better penalty for couples is ...
         informationBounds.Add(-1.0 * mantra.couples / mantra.bit_count);

      // Unless very informative marker pairs are present
      for (int i = 0; i < informativeCount - 1; i++)
         if (informationBounds[i] + informationBounds[i + 1] > 1.0)
            useSparse = true;
      }

#ifdef __PRINTOUT_STATS__
   int median = nodeCounts.dim / 2;
   int ciLo = nodeCounts.dim * 25 / 1000;
   int ciHi = nodeCounts.dim * 975 / 1000;

   char * labels [] = { "Nodes", "Leafs", "Zeros", "Info" };
   Vector * vect [] = { &nodeCounts, &leafCounts, &zeroCounts, &infoScores};

   for (int i = 0; i < 4; i++)
      {
      Vector & vector = *vect[i];

      vector.Sort();
      PrintMessage("  % 8s: %.2f (%.2f to %.2f) Median = %.2f, 95%% CI = %.2f - %.2f",
         labels[i], vector.Average(), vector.Min(), vector.Max(),
         vector[median], vector[ciLo], vector[ciHi]);
      }
#endif
   }

bool FamilyAnalysis::ScoreConditionals()
   {
   if (twopoint) return true;

   right = new Tree[informativeCount];
   rightScale.Dimension(informativeCount);
   leftScale.Dimension(informativeCount);

   Multipoint engine;

   if (useSparse) engine.SetupConditioning(mantra);

   int last = informativeCount - 1;

   rightScale[last] = 1.0;
   singlepoint[last].UnPack();
   right[last].Copy(singlepoint[last]);
   singlepoint[last].Discard();

   for (int m = last - 1; m >= 0; m--)
      {
      ProgressReport("Right Conditional", last - m, informativeCount);

      // Calculate recombination fraction
      double distance = informativePositions[m+1]-informativePositions[m];

      // Retrieve inheritance at m+1 given m + 1, m + 2, ...
      engine.Clear();
      engine.Copy(right[m+1]);
      right[m+1].Pack();

      if (useSparse && (informationBounds[m+1] + informationBounds[m] > 1.0))
         {
         // Try fast conditioning for informative locations?
         singlepoint[m].UnPack();
         right[m].Copy(singlepoint[m]);
         singlepoint[m].Discard();

         // Calculate conditional probability without intermediates
         engine.Condition(right[m], distance, rightScale[m+1]);
         }
      else
         {
         // Calculate inheritance at m given m+1, m+2, ...
         engine.MoveAlong(mantra, distance, rightScale[m+1]);

         // Calculate inheritance at m given *** m ***, m + 1, ...
         singlepoint[m].UnPack();
         engine.Multiply(singlepoint[m]);
         singlepoint[m].Discard();

         // Store it for future use
         right[m].Copy(engine);
         }

      // Check that the likelihod hasn't hit zero
      if (stats.GetMean(right[m]) == 0.0) return false;

      // If the tree is very unlikely, multiply by a constant to avoid underflow
      rightScale[m] = stats.mean < lowBound ? rescale : 1.0;
      }

   right[0].Pack();
   return true;
   }

void FamilyAnalysis::WalkLeft
   (Multipoint & left, Multipoint & leftMarker, Multipoint & full, int marker)
   {
   // Load required information to memory
   singlepoint[marker+1].UnPack();
   right[marker+1].UnPack();

   // Calculate the new left conditioned probabilities
   if (marker != -1)
      {
      // First, check how far from the previous location
      double distance = informativePositions[marker + 1] -
                        informativePositions[marker];

      // Load previous left conditioned likelihoods
      left.Copy(leftMarker);

      // Load the singlepoint likelihoods at the next location
      leftMarker.Copy(singlepoint[marker + 1]);

      if (!findErrors && useSparse &&
         (informationBounds[marker] + informationBounds[marker + 1] > 1.0))
         {
         // Build left conditioned probability up to marker + 1 (inclusive)
         left.Condition(leftMarker, distance, leftScale[marker]);
         leftScale[marker + 1] = stats.GetMean(leftMarker) < lowBound ? rescale : 1.0;

         // TODO -- We could choose to condition on left / right ?
         // Get right conditioned probability
         full.Copy(right[marker + 1]);
         left.Condition(full, distance, leftScale[marker]);
         }
      else
         {
          // Do the multipoint calculation
         left.MoveAlong(mantra, distance, leftScale[marker]);

         // Get the full left conditioned probability
         leftMarker.Multiply(left);
         leftScale[marker + 1] = stats.GetMean(leftMarker) < lowBound ? rescale : 1.0;

         // Get the full left and right conditioned probability
         full.Copy(right[marker + 1]);
         full.Multiply(left);
         }
       }
   else
      {
      // This is for the first marker
      leftMarker.Copy(singlepoint[0]);
      full.Copy(right[0]);
      leftScale[0] = 1.0;
      }

   // Check the marker for errors
   if (findErrors) ErrorCheck(left, full, marker + 1);

   // Free up memory
   singlepoint[marker+1].Discard();
   right[marker+1].Discard();
   }

void FamilyAnalysis::ScanChromosome()
   {
   // Anchors for starting position
   int leftAnchor  = -1;
   int rightAnchor =  0;
   int positionCount = analysisPositions.dim;

   // Left-conditioned probabilities up to left-marker
   Multipoint leftMarker;

   // Left and right conditionals up to current position
   Multipoint leftInheritance, inheritance;

   // Setup auxiliary variables for processing very sparse trees
   if (useSparse) leftInheritance.SetupConditioning(mantra);

   for (int i = 0; i < positionCount; i++)
      {
      ProgressReport("Scanning Chromosome", i, positionCount);

      double position = analysisPositions[i];

      // Move anchors so they flank the current position
      while (!zeroRecombination && !twopoint &&
             rightAnchor < informativeCount &&
             informativePositions[rightAnchor] <= position)
         {
         WalkLeft(leftInheritance, leftMarker, inheritance, leftAnchor++);
         rightAnchor++;
         }

      // Special case of two point analysis
      if (twopoint)
         {
         rightAnchor = leftAnchor = i;
         singlepoint[i].UnPack();
         inheritance.Copy(singlepoint[i]);
         singlepoint[i].Discard();
         }
      // If we are analysing at marker, the hard work is done
      else if (leftAnchor != -1 && informativePositions[leftAnchor] == position)
          {
          // nothing to do!
          }
      else if (!betweenMarkers)
         continue;
      // If there are no left flanking informative markers
      else if (leftAnchor == -1)
         {
         double distance = informativePositions[rightAnchor] - position;

         right[rightAnchor].UnPack();
         inheritance.Copy(right[rightAnchor]);
         inheritance.MoveAlong(mantra, distance);
         right[rightAnchor].Discard();
         }
      // If there are no right flanking informative markers
      else if (rightAnchor == informativeCount)
         {
         double distance = position - informativePositions[leftAnchor];

         inheritance.Copy(leftMarker);
         inheritance.MoveAlong(mantra, distance);
         }
      // Otherwise we are really between markers
      else
         {
         double distance = position - informativePositions[leftAnchor];

         leftInheritance.Copy(leftMarker);
         leftInheritance.MoveAlong(mantra, distance, leftScale[leftAnchor]);

         distance = informativePositions[rightAnchor] - position;

         right[rightAnchor].UnPack();
         inheritance.Copy(right[rightAnchor]);
         inheritance.MoveAlong(mantra, distance, rightScale[rightAnchor]);
         inheritance.Multiply(leftInheritance);
         right[rightAnchor].Discard();
         }

      // The probability of the data summed all possible inheritance vectors
      double lk;

      if (estimateMatrices) lk = matrix.Calculate(inheritance, labels[i]);
      if (estimateIBD) lk = ibd.Calculate(inheritance, labels[i]);
      if (estimateKinship || storeKinship)
         lk = kinship.Calculate(inheritance,
                                estimateKinship ? (const char *) labels[i] : NULL,
                                storeKinship ? i : -1);
      if (!estimateIBD && !estimateKinship && !estimateMatrices && !storeKinship)
         lk = stats.GetMean(inheritance);

      for (int a = 0; a < nplCount; a++)
         // Only analyse informative families (eg. multiple related affecteds)
         if (nplInv[a] != 0.0)
            kac.scores[a][serial][i] =
               (inheritance.MeanProduct(npl[a])/lk - nplMean[a]) * nplInv[a];

      if (estimateInformation)
         {
         // Calculate information
         stats.GetTreeInfo(inheritance);

         information[0][i] += stats.information * mantra.bit_count;
         information[1][i] += mantra.bit_count;
         }
      }
   }

void FamilyAnalysis::ErrorCheck
     (Multipoint & left, Multipoint & full, int informativeMarker)
   {
   if (informativeCount <= 1) return;

   Multipoint      rightInheritance;
   InheritanceTree alternative;

   double singlepoint_likelihood= stats.GetMean(singlepoint[informativeMarker]);
   double multipoint_likelihood = stats.GetMean(full);

   int marker = informativeMarkers[informativeMarker];

   if (++informativeMarker != informativeCount)
      {
      double distance = informativePositions[informativeMarker] -
                        informativePositions[informativeMarker-1];
      right[informativeMarker].UnPack();
      rightInheritance.Copy(right[informativeMarker]);
      right[informativeMarker].Discard();
      rightInheritance.MoveAlong(mantra, distance, rightScale[informativeMarker]);
      if (informativeMarker != 1)
         rightInheritance.Multiply(left);
      }

   for (int error = family->first; error <= family->last; error++)
      if (ped[error].markers[marker].isKnown())
         {
         Alleles saved_genotype = ped[error].markers[marker];
         ped[error].markers[marker][0] = ped[error].markers[marker][1] = 0;

         alternative.ScoreVectors(mantra, marker);

         double alternative_likelihood = stats.GetMean(alternative);

         if (alternative_likelihood != singlepoint_likelihood)
            {
            double alternative_multipoint;

            if (informativeMarker == informativeCount)
               alternative_multipoint = alternative.MeanProduct(left);
            else
               alternative_multipoint = alternative.MeanProduct(rightInheritance);

            double score =
               (multipoint_likelihood / alternative_multipoint) /
               (singlepoint_likelihood / alternative_likelihood);

            if (score < 0.025)
               {
               PrintMessage("  %s genotype for individual %s is unlikely [%.3e]",
                           (const char *) ped.markerNames[marker],
                           (const char *) ped[error].pid, score);
               fprintf(errorfile, "%10s %10s %10s %#10.3g\n",
                       (const char *) family->famid,
                       (const char *) ped[error].pid,
                       (const char *) ped.markerNames[marker], score);
               }
            }

         ped[error].markers[marker] = saved_genotype;
         }
   }

// Output buffering functions
//

void FamilyAnalysis::PrintHeader()
   {
   printHeader = false;

   printf("Family: %5s - Founders: %-2d - Descendants: %-2d - Bits: %-2d\n",
          (const char *) family->famid, mantra.f, mantra.n - mantra.f,
          mantra.bit_count);
   }

void FamilyAnalysis::PrintMessage(const char * msg, ...)
   {
   if (printHeader) PrintHeader();

   va_list ap;

   va_start (ap, msg);
   vprintf(msg, ap);
   printf("\n");
   va_end(ap);
   }

void FamilyAnalysis::ProgressReport(const char * msg, int done, int total)
   {
   if (quietOutput) return;

   long check = 0x10000 >> mantra.bit_count;
   if (mantra.bit_count > 16 || done & check )
      {
      if (printHeader)
         printf("Family %s (%d bits) -",
               (const char *) family->famid, mantra.bit_count);
      printf("  %s: %3d%%%20s\r", msg, done * 100 / total, "");
      fflush(stdout);
      cleanOutput = true;
      }
   }

void FamilyAnalysis::CleanMessages()
   {
   if (cleanOutput)
      {
      printf("%70s\r", "");
      fflush(stdout);
      }

   if (!printHeader)
      printf("\n");
   }

void FamilyAnalysis::ShowInformation()
   {
   if (!estimateInformation) return;

   printf("\n%15s %6s\n", "Position", "Info");

   for (int i = 0; i < labels.Length(); i++)
      printf("%15s %6.4f\n", (const char *) labels[i],
             information[1][i] == 0.0 ? 0.0 :
             information[0][i] / information[1][i]);
   }

void FamilyAnalysis::ShowLODs()
   {
   if (nplCount) kac.Analyse();
   if (storeKinship) vc.Analyse(ped, kinship, labels);
   }

void FamilyAnalysis::OpenErrorFile()
   {
   errorfile = fopen("merlin.err", "wt");
   fprintf(errorfile, "%10s %10s %10s %10s\n",
      "FAMILY", "PERSON", "MARKER", "RATIO");
   }


void FamilyAnalysis::SetupFiles()
   {
   if (findErrors) OpenErrorFile();
   if (estimateMatrices) matrix.OpenFile();
   if (estimateKinship) kinship.OpenFile();
   if (estimateIBD) ibd.OpenFile();
   if (bestHaplotype || (sampledHaplotypes != 0) || allHaplotypes) haplo.OpenFile();
   if (simwalk2) hybrid.OpenFiles();
   if (useSwap) BasicTree::SetupSwap();
   }

void FamilyAnalysis::CloseFiles()
   {
   if (findErrors)
      {
      printf("Unlikely genotypes listed in file [merlin.err]\n");
      fclose(errorfile);
      }
   if (estimateMatrices) matrix.CloseFile();
   if (estimateKinship) kinship.CloseFile();
   if (estimateIBD) ibd.CloseFile();
   if (bestHaplotype || (sampledHaplotypes != 0) || allHaplotypes) haplo.CloseFile();
   if (simwalk2) hybrid.CloseFiles();
   if (useSwap) BasicTree::CloseSwap();
   }

void FamilyAnalysis::SortInformation(int start, int end)
   {
   if (start >= end) return;

   double average = (informationScores[informationIndex[start]] +
                     informationScores[informationIndex[end]]) * 0.5;

   int a = start, b = end;

   while (true)
      {
      while ((informationScores[informationIndex[a]] >= average) && a < b)
         a++;

      while ((informationScores[informationIndex[b]] <= average) && a < b)
         b--;

      if (a == b) break;

      informationIndex.Swap(a, b);
      }

   SortInformation(start, a == end ? a - 1 : a);
   SortInformation(a == start ? a + 1 : a, end);
   }

bool FamilyAnalysis::ZeroRecombination()
   {
   // Identify the most informative markers
   informationIndex.Dimension(informativeCount);
   informationIndex.SetSequence(0, 1);
   SortInformation(0, informativeCount - 1);

   // Combined tree includes all nodes that are non-zero in all trees
   Tree combined;

   // Allocate storage for final result
   right = new Tree[1];
   singlepoint[informationIndex[0]].UnPack();
   right[0].Copy(singlepoint[informationIndex[0]]);
   singlepoint[informationIndex[0]].Discard();

   // Locate inheritance vectors that are compatible with all markers
   for (int i = 1; i < informativeCount; i++)
      {
      // Incorporate next most informative marker
      combined.Copy(right[0]);
      singlepoint[informationIndex[i]].UnPack();
      combined.Multiply(singlepoint[informationIndex[i]]);
      singlepoint[informationIndex[i]].Discard();
      combined.TrimZeros();
      right[0].Copy(combined);

      // Check that is possible given zero recombination
      if (stats.GetMean(right[0]) == 0.0)
         return false;

      // Avoid underflow by multiplying by large rescale constant
      if (stats.mean < lowBound)
         right[0].Multiply(0, rescale);
      }

   right[0].Pack();
   return true;
   }
 
